package org.mercury_im.messenger.xmpp;

import org.jivesoftware.smack.AbstractXMPPConnection;
import org.jivesoftware.smack.ReconnectionListener;
import org.jivesoftware.smack.ReconnectionManager;
import org.jivesoftware.smack.SmackException;
import org.jivesoftware.smack.XMPPException;
import org.jivesoftware.smack.sasl.SASLErrorException;
import org.jivesoftware.smack.util.Async;
import org.jivesoftware.smackx.csi.ClientStateIndicationManager;
import org.mercury_im.messenger.ClientStateListener;
import org.mercury_im.messenger.data.repository.Repositories;
import org.mercury_im.messenger.entity.Account;
import org.mercury_im.messenger.store.RosterStoreBinder;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.inject.Inject;
import javax.inject.Singleton;

import io.reactivex.Observable;
import io.reactivex.disposables.CompositeDisposable;
import io.reactivex.disposables.Disposable;
import io.reactivex.schedulers.Schedulers;

@Singleton
public class MercuryConnectionManager implements ClientStateListener {

    private static final Logger LOGGER = Logger.getLogger("ConnectionManager");
    private static final XmppConnectionFactory connectionFactory = new XmppConnectionFactory();

    private final Map<UUID, MercuryConnection> connections = new ConcurrentHashMap<>();
    private final RosterStoreBinder rosterStoreBinder;

    private Repositories repositories;
    private CompositeDisposable disposable = new CompositeDisposable();

    @Inject
    public MercuryConnectionManager(RosterStoreBinder rosterStoreBinder) {
        this.rosterStoreBinder = rosterStoreBinder;
    }

    public List<MercuryConnection> getConnections() {
        return new ArrayList<>(connections.values());
    }

    public MercuryConnection getOrCreateConnection(Account account) {
        MercuryConnection connection = connections.get(account.getId());
        if (connection == null) {
            connection = new MercuryConnection(account, connectionFactory.createConnection(account));
            bindConnection(connection);
            connections.put(account.getId(), connection);
        }
        return connection;
    }

    public void bindConnection(MercuryConnection connection) {
        rosterStoreBinder.setRosterStoreOn(connection);
        ReconnectionManager.getInstanceFor((AbstractXMPPConnection) connection.getConnection())
                .addReconnectionListener(new ReconnectionListener() {
                    @Override
                    public void reconnectingIn(int seconds) {
                        LOGGER.log(Level.FINER, "Reconnecting connection " + connection.getAccount().getAddress() + " in " + seconds + " seconds.");
                    }

                    @Override
                    public void reconnectionFailed(Exception e) {
                        LOGGER.log(Level.WARNING, "Reconnection of connection " + connection.getAccount().getAddress() + " failed.", e);
                    }
                });
    }

    public void connectActiveAccounts() {
        Disposable d = repositories.getAccountRepository()
                .observeAllAccounts()
                .subscribeOn(Schedulers.io())
                .observeOn(Schedulers.newThread())
                .firstOrError()
                .flatMapObservable(Observable::fromIterable)
                .doOnNext(account -> Async.go(() -> {
                    MercuryConnection connection = getOrCreateConnection(account);
                    if (connection.getAccount().isEnabled()) {
                        try {
                            connection.login();
                        } catch (SASLErrorException e) {
                            LOGGER.log(Level.SEVERE, "Account Credentials incorrect", e);
                            ((AbstractXMPPConnection) connection.getConnection()).disconnect();
                            ReconnectionManager.getInstanceFor((AbstractXMPPConnection) connection.getConnection()).disableAutomaticReconnection();
                        } catch (IOException | InterruptedException | SmackException | XMPPException e) {
                            e.printStackTrace();
                        }
                    }
                }))
                .subscribe();
        disposable.add(d);
    }

    // CSI

    @Override
    public void onClientInForeground() {
        LOGGER.log(Level.INFO, "CSI: active");
        for (MercuryConnection connection : getConnections()) {
            tryCsiActive(connection);
        }
    }

    @Override
    public void onClientInBackground() {
        LOGGER.log(Level.INFO, "CSI: inactive");
        for (MercuryConnection connection : getConnections()) {
            tryCsiInactive(connection);
        }
    }

    private void tryCsiActive(MercuryConnection connection) {
        try {
            ClientStateIndicationManager.active(connection.getConnection());
        } catch (SmackException.NotConnectedException | InterruptedException e) {
            LOGGER.log(Level.WARNING, "Sending CSI state 'active' failed.", e);
        } catch (IllegalArgumentException e) {
            LOGGER.log(Level.WARNING, "CSI not supported by server.");
        }
    }

    private void tryCsiInactive(MercuryConnection connection) {
        try {
            ClientStateIndicationManager.inactive(connection.getConnection());
        } catch (SmackException.NotConnectedException | InterruptedException e) {
            LOGGER.log(Level.WARNING, "Sending CSI state 'inactive' failed.", e);
        } catch (IllegalArgumentException e) {
            LOGGER.log(Level.WARNING, "CSI not supported by server.");
        }
    }

    public void setRepositories(Repositories repositories) {
        this.repositories = repositories;
    }
}
